#!/usr/bin/env perl

#	Masinky
#	Copyright (C) 2007 Michal Vaner (vorner@ucw.cz)
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	he Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

use warnings;
use strict;

$|=1;

my %settings;
{#Basic info
	#get the dimensions
	( $settings{width}, $settings{height} ) = ( <> =~ /^I:dimensions\s+(\d+)\s+(\d+)/ ) or die "F:no dimensions specified\n";
	#get the config file
	( $settings{config} ) = ( <> =~ /^I:config\s+"([^"]+)"/ ) or die "F:no config file\n";
}

{#Config
	open( CONFIG, $settings{config} ) or die "F:failed to open config\n";
	print "D:opened config, skipping sections up to masinky section\n";
	my $ln;
	while( defined( $ln = <CONFIG> ) && ( $ln !~ /^\s*\[masinky\]\s*$/ ) ) {};

	while( defined( $ln = <CONFIG> ) && ( $ln !~ /^\s*\[.+\]\s*$/ ) ) {
		next if( $ln =~ /^\s*(|#.*)$/ );
		my ( $name, $value ) = ( $ln =~ /^\s*([^ \t=]+)\s*=\s*(.*\S)\s*$/ ) or die "F:syntax error in config\n";
		$value =~ s/"(.*)"/$1/;#eat quotes if needed
		$settings{"C_$name"} = $value;
		print "D:reading value of $name:$value\n";
	}

	close CONFIG;

	foreach( ( "C_wagondefs", "C_traindefs", "C_timetables", "C_environdef", "C_workroot" ) ) {#This must be present
		defined( $settings{$_} ) or die "F:Config value missing for $_\n";
	}

	print "Q[]:cd " . $settings{C_workroot} . "\n";
	die "F:unable to CD\n" unless( <> =~ /R\[\]:cd\s*\n/ );
}

#load wagon definitions
my %wagons;
{#Wagons
	my ( $haveWagon, $haveWagonObject ) = ( 0 );

	sub newWagonObject() {
		print "D:creating a wagon object\n";
		$haveWagonObject = 1;
		return { "C_animTime" => 1, imgs => [] };
	}

	sub insWagon( $ $ ) {
		print "D:New wagon, inserting the old one\n";
		my ( $wagon, $name ) = @_;
		if( defined( $wagon->{name} ) ) {
			foreach( ( "C_width", "C_height" ) ) {#Check all walues present
				defined( $wagon->{$_} ) or die "F:Vagon value missing for $_ in $wagon->{name}\n";
			}
			foreach my $obj ( @{$wagon->{objs}} ) {
				foreach( ( "C_x", "C_y" ) ) {
					defined( $obj->{$_} ) or die "F:Vagon object value missing for $_ in $wagon->{name}\n";
				}
			}
			$wagons{$wagon->{name}} = $wagon;
		}
		$haveWagon = 1;
		$haveWagonObject = 0;
		return { name => $name, objs => [] };
	}

	sub checkWagon() { die "F:No wagon started\n" unless( $haveWagon ); }

	sub checkWagonObject() {
		checkWagon;
		die "F:No wagon object started\n" unless( $haveWagonObject );
	}

	my $wagon;
	sub readWFile( $ ) {
		my( $fn ) = @_;
		open( WAGONS, $fn ) or die "F:can not open wagon definitions $fn\n";
		foreach( <WAGONS> ) {
			next if( /^\s*(#.*)?$/ );#ignore comments and empty lines
			my( $name, $value, $img, $vals, $newFN );
			if( ( $name ) = /^\s*>>\s*(\S+)\s*$/ ) {#new wagon
				$wagon = insWagon( $wagon, $name );
			} elsif( /^\s*%\s*$/ ) {#new object
				checkWagon;
				push( @{$wagon->{objs}}, newWagonObject );
			} elsif( ( $name, $value ) = /^\s*!\s*([^ \t=]+)\s*=\s*(\d+)\s*$/ ) {#some value
				checkWagon;
				$wagon->{"C_$name"} = $value;
			} elsif( ( $img ) = /^\s*\*\s*(.*\S)\s*$/ ) {#image name
				checkWagonObject;
				push(@{$wagon->{objs}->[-1]->{imgs}}, $img);
			} elsif( ( $name, $vals ) = /^\s*>\s*([^ \t=]+)\s*=\s*(.*\S)\s*$/ ) {#some image value
				print "D:new image value\n";
				my( $val, $num );
				checkWagonObject;
				if( ( $val ) = ( $vals =~ /^\([^,]\)$/ ) ) {
					$vals = $val;
				}
				$wagon->{objs}->[-1]->{"C_$name"} = $vals;
			} elsif( ( $newFN ) = /^\s*@\s*(.*\S)\s*$/ ) {#Source a file
				my $orig = $fn;
				$orig =~ s/[^\/]+$//;
				&readWFile( $orig . $newFN );
			} else {
				die "F:syntax error at wagon definitions: $_\n";
			}
		}
		close( WAGONS );
	}
	readWFile( $settings{C_wagondefs} );
	insWagon( $wagon, undef );
}

#Push the wagon definitions/images to the server
for ( values( %wagons ) ) {
	my $name = "wagon".$_->{name};
	my( $lname, $rname ) = ( "left_$name", "right_$name" );
	print "M:def $lname\nM:def $rname\nM:placeObj $rname $lname obj 0 0 0\nM:flipX $rname\n";
	my $z = scalar( @{$_->{objs}} );
	for my $img ( @{$_->{objs}} ) {
		my $set = 0;
		for my $pic ( @{$img->{imgs}} ) {
			if( $set ) {
				print "M:addImg tmp_img $pic " . int( 1000 * $img->{C_animTime} ) . "\n";
			} else {
				$set = 1;
				print "M:placeImg $lname $pic $pic $img->{C_x} $img->{C_y} " . $z -- . " tmp_img\n";
			}
		}
	}
}

#Load train definitions
my %traindefs;
{
	my( $train, $trainname );
	$train = [];

	sub putTrain( $ ) {
		my( $name ) = @_;
		if( defined $trainname ) {
			print "D:inserting train $trainname\n";
			$traindefs{$trainname} = $train;
		}
		$train = [];
		$trainname = $name;
	}

	sub grabWagon( $$ ) {
		my( $wagname, $probab ) = @_;
		defined( $wagons{$wagname} ) or die "F:no vagon of that name\n";
		return { wagon => $wagons{$wagname}, probab => $probab };
	}

	sub loadTrains( $ ) {
		my( $fn ) = @_;
		print "D:train defs $fn\n";
		open( TRAINS, $fn ) or die "F:Could not open train definitions $fn\n";
		foreach( <TRAINS> ) {
			next if( /^\s*(#.*|)$/ );#Skip comments
			if( my( $name ) = /^\s*\[\s*(.*\S)\s*\]\s*$/ ) {
				putTrain( $name );
			} elsif( my( $filename ) = /^\s*@\s*(.*\S)\s*$/ ) {
				my $orig = $fn;
				$orig =~ s/[^\/]*$//;
				&loadTrains( $orig . $filename );
			} elsif( my( $obj ) = /^\s*\*\s*(.*\S)\s*$/ ) {
				my( $modifs, $wagondf ) = ( " $obj" =~ /^(.*)\s(\[.*\]|\(.*\)|\S*)\s*$/ );#Get the modificators and wagon definition
				my $wagon;
				if( my( $tname ) = ( $wagondf =~ /^\s*\[\s*(.*\S)\s*\]\s*$/ ) ) {#Embed a train
					die "F:Unkown train name to embed $tname\n" unless( defined $traindefs{$tname} );
					print "D:embed train $tname\n";
					$wagon = {
						type => "train",
						value => $traindefs{$tname}
					};
				} elsif( my( $waglist ) = ( $wagondf =~ /^\s*\(\s*(.*\S)\s*\)\s*$/ ) ) {#Choose from more wagons
					my @wagons;
					print "D:inserting a choise wagon\n";
					foreach( split( /\|/, $waglist ) ) {
						chomp;
						my( $wname, $probab ) = /^(.+)\/(\d+)$/;
						( $wname, $probab ) = ( $_, 1 ) unless( defined $probab );
						push @wagons, grabWagon( $wname, $probab );
					}
					$wagon = {
						type => "choise",
						value => \@wagons
					};
				} else {#Just a wagon
					print "D:insert wagon\n";
					$wagon = {
						type => "choise",
						value => [ grabWagon( $wagondf, 1 ) ]
					};
				}
				my $modif;
				while( ( $modif, $obj ) = ( "$obj " =~ /^\s*(\<\s*\d+\s*,\s*\d+\s*\>|%\s*\d+)\s(.*)$/ ) ) {
					print "D:wagon modifier $modif\n";
					if( my( $probab ) = ( $modif =~ /^%\s*(\d+)\s*$/ ) ) {#Probability
						$wagon->{probab} = $probab * 0.01;
					} elsif( my( $min, $max ) = ( $modif =~ /^\<\s*(\d+)\s*,\s*(\d+)\s*\>/ ) ) {#range
						( $wagon->{min}, $wagon->{max}, $wagon->{hasrange} ) = ( $min, $max, 1 );
					} else {
						die "F:Unknown wagon modifier in $fn\n";#Can Not Happen
					}
				}
				push @{$train}, $wagon;
			} else {
				die "F:Invalid syntax in train definitions\n";
			}
		}
		close TRAINS;
	}

	loadTrains( $settings{C_traindefs} );#Load all the trains
	putTrain( undef );#Push there the last one
}

#Environment and stations
#sub loadEnviron( $ );

my $envpos = 0;

sub loadEnviron( $ ) {
	my( $env ) = @_;
	my( %stations );
	print "D:loading environment defs $env\n";
	open( ENV, $env ) or die "F:could not open env file $env\n";
	sub perc( $ $ ) {
		my( $val, $total ) = @_;
		my( $num, $inc );
		if( ( $num, $inc ) = ( $val =~ /(\d+)%([+-]\d+|)/ ) ) {
			$num = $total * $num / 100;
			if( $inc ne "" ) {
				$num += $inc;
			}
			return $num;
		} else {
			return $val;
		}
	}
	foreach( <ENV> ) {
		my( $filename, $imgname, $x, $y, $z, $name, $color, $width, $x1, $x2, $y1, $y2 );
		next if ( /^\s*(#.*|)$/ );#Skip comments and empty lines
		if( ( $filename ) = /^\s*@\s*(.*\S)\s*$/ ) {#Include file
			&loadEnviron( $filename );
		} elsif( ( $imgname, $x, $y, $z ) = /^\s*\*\s*img\s+(\S+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+)\s*$/ ) {#Image
			$x = perc( $x, $settings{"width"} );#Compute the values
			$y = perc( $y, $settings{"height"} );
			print "D:new environment object at $x $y\n";
			print "M:placeImg root $imgname envImage $x $y $z\n";#Draw it
		} elsif( ( $color, $width, $x1, $y1, $x2, $y2, $z ) = /^\s*\*\s*line\s+([0-9a-fA-F]{6})\s+(\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+)\s*$/ ) {
			$x1 = perc( $x1, $settings{"width"} );
			$x2 = perc( $x2, $settings{"width"} );
			$y1 = perc( $y1, $settings{"height"} );
			$y2 = perc( $y2, $settings{"height"} );
			print "D:new environment line at $x1 $y1 - $x2 $y2\n";
			print "M:placeLn root $color $width envLn $x1 $y1 $x2 $y2 $z\n";
		} elsif( ( $name, $x, $y, $z ) = /^\s*_\s*(\S+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+%?|\d+%[+-]\d+)\s+(\d+)$/ ) {#Station
			$x = perc( $x, $settings{"width"} );
			$y = perc( $y, $settings{"height"} );
			print "D:new station at $x $y\n";
			$stations{$name} = {
				x => $x,
				y => $y,
				z => $z,
				name => $name
			};
		} else {
			die "F:invalid syntax in $env\n";
		}
	}
	close ENV;
	return \%stations;
}

my %stations = %{loadEnviron( $settings{C_environdef} )};

sub minutes( $$ ) {
	my( $hour, $minute ) = @_;
	return 60 * $hour + $minute;
}

my @timetables;

{#Load the tables
	my( $actline, $stationname, $laststation, $lasttrack, $waittime, $speed, $priority, $direction, $starthour, $endhour, $startminute, $endminute, $restrictions, $lockval );

	sub checkNonEmpty() {
		if( $actline && ( ( @{$actline->{stations}} <= 1 ) || ( ! @{$actline->{entries}} ) ) ) {#No empty tracks please
			print "D:Empty line, leaving out ($actline->{name})\n";
			pop( @timetables );
		}
	}

	sub loadTable( $ ) {
		my( $fn ) = @_;
		open( TABLE, $fn ) or die "F:could not open timetable $fn\n";

		foreach( <TABLE> ) {
			next if( /^\s*(|#.*)$/ );#comments
			my( $incfile, $linename, $trainname );
			if( ( $incfile ) = /^\s*@\s*(.*\S)\s*$/ ) {#includes
				&loadTable( $incfile );
			} elsif( ( $linename ) = /^\s*\[\s*(.*\S)\s*\]\s*$/ ) {#Start the train line
				print "D:New timetable line $linename\n";
				checkNonEmpty();
				push( @timetables, ( $actline = {
					name => $linename,
					stations => [],
					entries => []
				} ) );
				$laststation = $lasttrack = undef;
			} elsif( ( $trainname ) = /^\s*\$\s*(.*\S)\s*$/ ) {
				die "F:unknown train $trainname\n" unless defined $traindefs{$trainname};
				die "F:no line started\n" unless defined $actline;
				print "D:setting the train for line $actline->{name} to $trainname\n";
				$actline->{train} = $traindefs{$trainname};
			} elsif( ( $stationname ) = /^\s*\*\s*(\S+)\s*$/ ) {#New station in the list
				die "F:unknown station $stationname\n" unless defined $stations{$stationname};
				die "F:no line started\n" unless defined $actline;
				print "D:Adding a train station to the track: $stationname\n";
				$laststation->{forward} = $lasttrack;
				$laststation = {
					station => $stations{$stationname},
					back => $lasttrack,
					waittime => 0,
					backlock => 1,
					fwdlock => 1
				};
				my %nexttrack = (
					prev => $laststation,
					speed => 20,
				);
				if( defined $lasttrack ) {
					$lasttrack->{next} = $laststation;
				}
				$lasttrack = \%nexttrack;
				push @{$actline->{stations}}, $laststation;
			} elsif( ( $waittime ) = /^\s*\.\s*(\d+)\s*$/ ) {
				die "F:no station to setup\n" unless( defined $laststation );
				$laststation->{waittime} = $waittime;
			} elsif( ( $lockval ) = /^\s*=>\s*(\d+)\s*$/ ) {
				$laststation->{fwdlock} = $lockval;
			} elsif( ( $lockval ) = /^\s*<=\s*(\d+)\s*$/ ) {
				$laststation->{backlock} = $lockval;
			} elsif( ( $speed ) = /^\s*\|\s*(\d+)\s*$/ ) {
				die "F:no track to setup\n" unless( defined $lasttrack );
				$lasttrack->{speed} = $speed;
			} elsif( ( $direction, $priority ) = /^\s*([><])\s*\*\s*(\d+)\s*$/ ) {
				die "F:no line started\n" unless( defined $actline );
				print "D:New default schedule entry\n";
				push @{$actline->{entries}}, {
					anytime => 1,
					forward => ( $direction eq ">" ),
					priority => $priority
				};
			} elsif( ( $direction, $starthour, $startminute, $endhour, $endminute, $priority, $restrictions ) = /^\s*([><])\s*(\d\d):(\d\d)-(\d\d):(\d\d)\s+(\d+)((\s+@\s*(\d|(\d\d?|\*)-(\d\d?|\*)))*)\s*$/ ) {
				die "F:no line started\n" unless( defined $actline );
				print "D:New schedule entry\n";
				my %entry = (
					starttime => minutes( $starthour, $startminute ),
					endtime => minutes( $endhour, $endminute ),
					forward => ( $direction eq ">" ),
					priority => $priority,
					restrictions => [],
					anytime => 0,
					laststart => 0
				);
				$entry{wraparound} = ( $entry{starttime} > $entry{endtime} );
				$entry{intlen} = $entry{wraparound} ? ( 24 * 60 - ( $entry{starttime} - $entry{endtime} ) ) : ( $entry{endtime} - $entry{starttime} );
				my $restriction;
				$restrictions = $restrictions . " ";
				while( ( $restriction ) = ( $restrictions =~ /\s*@\s*(\d|(\d\d?|\*)-(\d\d?|\*))\s+/ ) ) {
					$restrictions =~ s/\s*@\s*(\d|(\d\d?|\*)-(\d\d?|\*))//;
					print "D: restriction $restriction | $restrictions\n";
					unless( $restriction =~ /-/ ) {
						push @{$entry{restrictions}}, {
							type => "day",
							day => $restriction
						};
					} else {
						my( $day, $month ) = ( $restriction =~ /(\d+|\*)-(\d+|\*)/ );
						push @{$entry{restrictions}}, {
							type => "date",
							day => $day,
							month => $month
						}
					}
				}
				push @{$actline->{entries}}, \%entry;
			} else {
				die "F:unknown statement in time tables $_\n";
			}
		}
		close( TABLE );
	}
	loadTable( $settings{C_timetables} );
	checkNonEmpty();
	#Postprocess the time tables
	foreach my $table ( @timetables ) {
		my( $actst, $actnode, $lastst );
		for( $actst = $table->{stations}[ 0 ]; defined( $actst ); $lastst = $actst, $actst = ( defined( $actst->{forward} ) ? $actst->{forward}->{next} : undef ) ) {#The forward ones
			my $new = {
				station => $actst->{station},
				waittime => $actst->{waittime},
				next => undef,
				speed => defined( $actst->{forward} ) ? $actst->{forward}->{speed} : 0,
				lockcnt => $actst->{fwdlock}
			};
			if( defined( $actnode ) ) {
				$actnode->{next} = $new;
			} else {
				$table->{forward} = $new;
			}
			$actnode = $new;
		}
		$actnode = undef;
		for( $actst = $lastst; defined( $actst ); $actst = ( defined( $actst->{back} ) ? $actst->{back}->{prev} : undef ) ) {#The backward ones
			my $new = {
				station => $actst->{station},
				waittime => $actst->{waittime},
				next => undef,
				speed => defined( $actst->{back} ) ? $actst->{back}->{speed} : 0,
				lockcnt => $actst->{backlock}
			};
			if( defined( $actnode ) ) {
				$actnode->{next} = $new;
			} else {
				$table->{backward} = $new;
			}
			$actnode = $new;
		}
		delete( $table->{stations} );
		sub genAscends( $ ) {
			my( $last ) = @_;
			my $act = $last->{next};
			while( defined $act ) {
				$last->{ascend} = ( ( $act->{station}->{x} != $last->{station}->{x} ) ? ( $act->{station}->{y} - $last->{station}->{y} ) / ( $act->{station}->{x} - $last->{station}->{x} ) : 0 );
				$last = $act;
				$act = $act->{next};
			}
			$last->{ascend} = 0;
		}
		genAscends( $table->{forward} );
		genAscends( $table->{backward} );
	}
}

#So, all loaded, start the usual run

my( %req, %events, @eventQueue );
my $lastId = 0;
my @actTrains;

sub hookEvent( $$$$ ) {
	my( $hookCmd, $sub, $single, $tag ) = @_;
	my $id = $lastId ++;
	$events{$id} = {
		hookCmd => $hookCmd,
		code => $sub,
		single => $single,
		tag => $tag
	};
	print "D:Hooking an event: $id/$hookCmd\n";
	print "H[$id]:$hookCmd\n";
}

sub request( $ ) {
	my( $reqCmd ) = @_;
	my( $id ) = $lastId ++;
	print "D:request $id/$reqCmd\n";
	print "Q[$id]:$reqCmd\n";
	until( readInp() eq "req" ) {};#Process it until you find the response to the request
	die( "F:asyncronous response? $req{line}\n" ) if( $id ne $req{id} );
	return( $req{line} );
}

sub genNewTick( $ );

sub tryLock( $$$ ) {#Starting station node (in the track), already locked count by me, wanted to lock
	my( $station, $lckcount, $wanted ) = @_;
	for( my( $i, $actstation ) = ( 0, $station ); ( $i < $wanted ) && ( $actstation ); $i ++, $actstation = $actstation->{next} ) {
		next if( $i < $lckcount );#Skip over the ones already locked by me
		return 0 if( $actstation->{station}->{locked} );#This is already locked, can not continue
	}
	#Ok, it can be locked, so lock it
	for( my( $i, $actstation ) = ( 0, $station ); ( $i < $wanted ) && ( $actstation ); $i ++, $actstation = $actstation->{next} ) {
		$actstation->{station}->{locked} = 1;
	}
	return 1;
}

sub populateWagons( $$$ ) {
	my( $dstref, $trainref, $depth ) = @_;
	die "F:recursive definition of trains\n" if( $depth > 10000 );#No-one is going to make such long trains
	foreach( @{$trainref} ) {
		#First, check the modifiers
		my( $probab, $mincnt, $maxcnt ) = ( 1, 1, 1 );#The defaults
		$probab = $_->{probab} if( defined( $_->{probab} ) );#Overwrite with provided values if any
		( $mincnt, $maxcnt ) = ( $_->{min}, $_->{max} ) if( $_->{hasrange} );
		next if( rand > $probab );#Will we use this generator?
		my $cnt = $mincnt + int( rand( $maxcnt - $mincnt + 1 ) );
		for( my $i = 0; $i < $cnt; ++ $i ) {#How many times do we call this one?
			if( $_->{type} eq "train" ) {#Include other train here
				&populateWagons( $dstref, $_->{value}, $depth + 1 );
			} elsif( $_->{type} eq "choise" ) {#Choose one of the possibilities
				my $sum = 0;
				foreach my $p ( @{$_->{value}} ) { $sum += $p->{probab} };
				my $shot = rand( $sum );
				foreach my $p ( @{$_->{value}} ) {
					if( $shot <= $p->{probab} ) {#Got it
						push @{$dstref}, $p->{wagon};
						last;
					} else {
						$shot -= $p->{probab};
					}
				}
			} else {
				die "F:Can Not Happen: unknown wagon generator type\n";#Can Not Happen
			}
		}
	}
}

sub wagonInStation( $$ ) {
	my( $trash, $data ) = @_;
	my( $train, $windex, $station ) = ( $data->{train}, $data->{windex}, $data->{station} );
	return if( $train->{deleted} );
	hookEvent( "position root $train->{obj_prefix}X " . ( $station->{station}->{x} + $train->{wagposs}->[$windex + 1] * $train->{dirmult} * -1 ) . " 0 $train->{hookmark} ?", \&wagonInStation, 1, {
		windex => $windex + 1,
		train => $train,
		station => $station
	} ) if( $windex < $train->{lastwagon} );
	#Set the vertical speed
	$train->{vertspeeds}->[ $windex ] = $station->{ascend};
	print "M:position $train->{obj_prefix}X $train->{obj_prefix}$windex 0 "
		. ( $train->{speed} * -1 * $train->{dirmult} * $train->{vertspeeds}->[ $windex ] ) . " 1\n"
		. "M:position $train->{obj_prefix}X $train->{obj_prefix}$windex " . ( $train->{wagposs}->[$windex] * $train->{dirmult} ) . " " . ( $station->{station}->{y} - $train->{wagons}->[$windex]->{C_height} ) . "\n";

}

sub delWagon( $$ ) {
	my( $trash, $data ) = @_;
	my( $train, $windex ) = ( $data->{train}, $data->{windex} );
	return if( $train->{deleted} );
	print "M:unplace $train->{obj_prefix}X $train->{obj_prefix}$windex\n";
	$train->{firstwagon} ++;
	$train->{tohide} --;
	return if( $train->{tohide} <= 1 );#Already hooked as the end of the train
	hookEvent( "position root $train->{obj_prefix}X " . ( $train->{headstation}->{station}->{x} + $train->{dirmult} * -1 * ( $train->{wagons}->[ $windex + 1 ]->{C_width} + $train->{wagposs}->[ $windex + 1 ] ) ) . " 0 $train->{hookmark} ?", \&delWagon, 1, {
		train => $train,
		windex => $windex + 1
	} );
}

sub startTrain( $$ );

sub enterStation( $$ ) {
	my( $trash, $train ) = @_;
	return if( $train->{deleted} );
	if( ( $train->{lastwaitstation} != $train->{headstation} ) && ( $train->{headstation}->{waittime} ) ) {
		#Hook the start after the time
		my( $timestamp ) = ( request( "timestamp" ) =~ /timestamp\s+(\d+)/ );
		hookEvent( "timestamp " . int( $timestamp + 1000 * $train->{headstation}->{waittime} ), \&startTrain, 1, $train );
		$train->{lastwaitstation} = $train->{headstation};
		#Stop the train
		print "M:position root $train->{obj_prefix}X 0 0 1\n";
		for( my $i = $train->{firstwagon}; $i <= $train->{lastwagon}; $i ++ ) {
			print "M:position $train->{obj_prefix}X $train->{obj_prefix}$i 0 0 1\n";
		}
		return;
	}
	if( defined $train->{headstation}->{next} && !tryLock( $train->{headstation}->{next}, -- $train->{lockcnt}, $train->{headstation}->{next}->{lockcnt} ) ) {#Couldn't lock, stop the train
		print "M:position root $train->{obj_prefix}X 0 0 1\n";
		for( my $i = $train->{firstwagon}; $i <= $train->{lastwagon}; $i ++ ) {
			print "M:position $train->{obj_prefix}X $train->{obj_prefix}$i 0 0 1\n";
		}
		$train->{waiting} = 1;
		return;
	}
	wagonInStation( undef, {
		windex => 0,
		train => $train,
		station => $train->{headstation}
	} );
	if( defined $train->{headstation}->{next} ) {#No speed changes and so on the last station
		$train->{speed} = $train->{headstation}->{speed};
		$train->{headstation} = $train->{headstation}->{next};
		my $speed = $train->{speed} * $train->{dirmult} * -1;
		print "M:position root $train->{obj_prefix}X $speed 0 1\n"
			. "M:setZ root $train->{obj_prefix}X $train->{headstation}->{station}->{z}\n";
		for( my $i = $train->{firstwagon}; $i <= $train->{lastwagon}; ++ $i ) {
			print "M:position $train->{obj_prefix}X $train->{obj_prefix}$i 0 " . ( $speed * $train->{vertspeeds}->[ $i ] ) . " 1\n";
		}
		hookEvent( "position root $train->{obj_prefix}X $train->{headstation}->{station}->{x} 0 $train->{hookmark} ?", \&enterStation, 1, $train );
	} else {
		hookEvent( "position root $train->{obj_prefix}X " . ( $train->{headstation}->{station}->{x} + $train->{dirmult} * -1 * $train->{wagons}->[ 0 ]->{C_width} ) . " 0 $train->{hookmark} ?", \&delWagon, 1, {
			train => $train,
			windex => 0
		} );
	}
}

sub startTrain( $$ ) {
	my( $trash, $train ) = @_;
	enterStation( undef, $train );
}

sub leaveStation( $$ ) {
	my( $trash, $train ) = @_;
	return if( $train->{deleted} );
	print "D:Train $train->{obj_prefix}: leave station\n";
	my( $station, $wagon );
	if( $train->{toshow} ) {#More wagons to show?
		$wagon = @{$train->{wagons}}[ ++ $train->{lastwagon} ];
		my $wagname = ( $train->{left} ? "left_" : "right_" ) . "wagon" . $wagon->{name};
		my $objname = $train->{obj_prefix} . $train->{lastwagon};
		$station = $train->{tailstation}->{station};
		push @{$train->{vertspeeds}}, 0;#Place there something, it will get overwritten soon
		print "M:placeObj $train->{obj_prefix}X $wagname $objname " . ( $train->{wagposs}->[$train->{lastwagon}] * $train->{dirmult} ) . " " . ( $station->{y} - $wagon->{C_height} ) . " 0\n";
		$train->{toshow} --;
		$train->{tohide} ++;
		wagonInStation( undef, {
			windex => $train->{lastwagon},
			train => $train,
			station => $train->{tailstation}
		} );
	} else {
		$train->{tailstation}->{station}->{locked} = 0;#Unlock
		#Check if anything can go now
		my @waiting = ();
		foreach( @actTrains ) {
			push @waiting, $_ if( $_->{waiting} );
		}
		@waiting = sort { $b->{priority} <=> $a->{priority} } @waiting;
		foreach( @waiting ) {
			startTrain( undef, $_ );#Try start it again, it won't go if it can't
		}
		if( $train->{tailstation} == $train->{headstation} ) { #We leave the last station, remove the train from list
			my @new = ();
			foreach( @actTrains ) {
				push @new, $_ if( $_ != $train );
			}
			@actTrains = @new;
			#Remove the last wagon
			print "M:unplace $train->{obj_prefix}X $train->{obj_prefix}$train->{lastwagon}\n"
				. "M:unplace root $train->{obj_prefix}X\n"
				. "M:undef $train->{obj_prefix}X\n";
			$train->{deleted} = 1;
			return;#No hook next leave station
		} else {
			$station = ( $train->{tailstation} = $train->{tailstation}->{next} )->{station};#The next tail station
			$wagon = @{$train->{wagons}}[$train->{lastwagon}];
		}
	}
	hookEvent( "position root $train->{obj_prefix}X " . ( -1 * ( $wagon->{C_width} + $train->{wagposs}->[$train->{lastwagon}] ) * $train->{dirmult} + $station->{x} ) . " 0 $train->{hookmark} ?", \&leaveStation, 1, $train );#When we leave the next one
}

my $lastObjPrefix = 0;

sub genTrain( $$ ) {
	my( $match, $timestamp ) = @_;
	return unless( tryLock( $match->{station}, 0, $match->{station}->{lockcnt} + 1 ) );
	print "D:Creating train\n";
	my( @wagonlist );
	populateWagons( \@wagonlist, $match->{line}->{train}, 1 );
	die "F:Empty train\n" unless( @wagonlist );
	print "D:It will be " . scalar @wagonlist . " wagons long\n";
	my $pos = 0;
	my @poss;
	foreach( @wagonlist ) {
		push @poss, $pos;
		$pos += $_->{C_width};
	}
	my $left = $match->{station}->{station}->{x} > $match->{station}->{next}->{station}->{x};
	my $train = {
		wagons => \@wagonlist,
		obj_prefix => "train-" . ( $lastObjPrefix ++ ) . "-",
		lockcnt => $match->{station}->{lockcnt} + 1, #The one will be eaten right away in this station
		wagposs => \@poss,
		left => $left,
		dirmult => ( $left ? 1 : -1 ),
		headstation => $match->{station},
		tailstation => $match->{station},
		toshow => scalar @wagonlist,
		tohide => 0,
		firstwagon => 0,
		lastwagon => -1,
		speed => 0, #Will get set soon
		vertspeeds => [], #The vertical speed multipliers of the wagons
		waiting => 0, #Are we waiting to unlock some station to continue?
		hookmark => ( $left ? '<=' : '>=' ), #The mark to hook it
		lastwaitstation => 0,#No wait performed
		priority => $match->{entry}->{priority},
		deleted => 0
	};
	print "M:def $train->{obj_prefix}X\nM:placeObj root $train->{obj_prefix}X $train->{obj_prefix}X $match->{station}->{station}->{x} 0 $match->{station}->{station}->{z}\n";
	leaveStation( undef, $train );
	startTrain( undef, $train );#The speed and so on
	$match->{entry}->{laststart} = $timestamp;#Just started
	return $train;
}

sub tickHook() {
	my( $timestamp ) = ( request( "timestamp" ) =~ /timestamp\s+(\d+)/ );
	my( $year, $month, $day, $hour, $minute ) = ( request( "time" ) =~ /time\s+(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2})/ );
	my( $weekday ) = ( request( "weekday" ) =~ /weekday\s+(\d)/ );
	print "D:started scanning the tables\n";
	my $actMinutes = minutes( $hour, $minute );
	my @matched = ();
	foreach( @timetables ) {#Scan which trains can go according to the time table
		print "D:$_->{name}\n";
		foreach my $entry ( @{$_->{entries}} ) {
			unless( $entry->{anytime} ) {#Abort if it does not match the table
				my( $inInterval, $dayDrift ) = ( 0, 0 );
				if( $entry->{wraparound} ) {
					$inInterval = ( ( $entry->{starttime} <= $actMinutes ) || #Between start and midnight
							( $entry->{endtime} >= $actMinutes ) );#Between midnight and the end
					$dayDrift = ( $entry->{endtime} >= $actMinutes );#Is it the next day?
				} else {
					$inInterval = ( ( $entry->{starttime} <= $actMinutes ) && #Started
							( $entry->{endtime} >= $actMinutes ) );#Did not yet ended
				}
				next unless( $inInterval );#Sorry, not now...
				next if( $entry->{laststart} + ( $entry->{intlen} * 60 * 1000 ) >= $timestamp );#It already started during this interval
				my $restrict = 0;
				print "D:To restricts\n";
				foreach my $restriction ( @{$entry->{restrictions}} ) { #Are there any other restrictions to take care of?
					if( $restriction->{type} eq "day" ) {
						my $wanted = $restriction->{day};
						if( $dayDrift ) {
							$wanted ++;
							$wanted = 0 if( $wanted == 7 );
						}
						print "D:day restriction $wanted ?= $weekday\n";
						unless( $wanted == $weekday ) {
							$restrict = 1;
							last;
						}
					} elsif( $restriction->{type} eq "date" ) {
						my( $dayWanted, $monthWanted ) = ( $restriction->{day}, $restriction->{month} );
						if( $dayDrift ) {
							sub isLeapYear( $ ) {
								my( $year ) = @_;
								return( ( $year % 4 == 0 ) && ( ( $year % 100 != 0 ) || ( $year % 400 == 0 ) ) );
							}
							my @dcounts = (
								31,
								( isLeapYear( $year ) ? 29 : 28 ),
								31,
								30,
								31,
								30,
								31,
								31,
								30,
								31,
								30,
								31 );
							unless( $dayWanted eq "*" ) {
								$dayWanted ++;
								if( $dayWanted > $dcounts[ $monthWanted ] ) {
									$dayWanted = 0;
									$monthWanted = 1 if( $monthWanted ++ > 12 );
								}
							}
						}
						print "D:date restriction check $dayWanted $monthWanted $day $month\n";
						unless( ( ( $dayWanted eq "*" ) || ( $dayWanted == $day ) ) &&
								( ( $monthWanted eq "*" ) || ( $monthWanted == $month ) ) ) {
							$restrict = 1;
							last;
						}
					} else {
						die "F:Can Not Happen: unknown restriction type";
					}
				}
				next if( $restrict );#No, not now
			}
			print "D:train match on $_->{name}\n";
			push @matched, {
				entry => $entry,
				priority => $entry->{priority},
				station => ( $entry->{forward} ? $_->{forward} : $_->{backward} ),
				line => $_
			};
		}
	}
	my @tmp;#Generate random order of requests
	while( @matched ) { push( @tmp, splice( @matched, int( rand( scalar( @matched ) + 1 ) ), 1 ) ); }
	#sort according the priority, highest first
	@matched = sort { $b->{priority} cmp $a->{priority} } @tmp;
	my $possible;
	print "D:matched " . scalar @matched . "\n";
	foreach $possible ( @matched ) {
		my $train = genTrain( $possible, $timestamp );
		next if( !defined( $train ) );#It couldn't be done
		push @actTrains, $train;
	}
	genNewTick( $timestamp );
}

sub genNewTick( $ ) {
	my( $timestamp ) = @_;
	$timestamp += 15000;#Quarter a minute later, please
	hookEvent( "timestamp $timestamp", \&tickHook, 1, undef );
}

tickHook();

sub readInp() {
	$_ = <STDIN>;
	return( undef ) unless( defined( $_ ) );
	return( "none" ) if( /^\s*$/ );
	my( $id, $line );
	if( ( $id, $line ) = /E\[([^\[\]]+)\]:(.*)$/ ) {#An event
		if( defined( $events{$id} ) ) {#Does such hook exist?
			my %event = %{$events{$id}};
			$event{line} = $line;
			push( @eventQueue, \%event );#Queue the event for execution
			delete( $events{$id} ) if( $event{single} );#Was it one-happen event?
		} else {
			die( "F:unknown event $id\n" );
		}
		return( "event" );
	} elsif( ( $id, $line ) = /R\[([^\[\]]+)\]:(.*)$/ ) {#A request
		$req{id} = $id;
		$req{line} = $line;
		return( "req" );
	} else {
		die( "F:Unknown command $_\n" );
	}
}

sub processEvents() {
	while( @eventQueue ) {
		my %event = %{shift( @eventQueue )};
		&{$event{code}}( $event{line}, $event{tag} );
	}
}

while( defined( readInp() ) ) {#Process lines until the end of the times
	processEvents();
}
